//=============================================================================
//  2023 msfp
//  MIT ExpandChordSymbols.js https://github.com/markshepherd/ExpandChordSymbols 
//  AniMikatamon started Chord Identifier (Pop & Jazz) plugin
//    https://musescore.org/en/node/311049
//    as an improvement of Chord Identifier Special V3 by Ziya Mete Demircan
//    https://musescore.org/en/node/287248
//    who was also very helpful in commenting and checking on the progress.
//  Copyright (C) 2016 Emmanuel Roussel - https://github.com/rousselmanu/msc_plugins
//  Emmanuel Roussel started his plugin as an improvement of the "Find Harmonies" plugin by Andresn 
//      (https://github.com/andresn/standard-notation-experiments/tree/master/MuseScore/plugins/findharmonies)
//      itself being an enhanced version of "findharmony" by Merte (https://musescore.org/en/node/7941)
//      It was ported to MuseScore 2.3 by Ke Xu and to MS 3 by Dmitri Ovodok
//  Emmanuel Roussel took some lines of code or got inspiration from:
//  - Berteh (https://github.com/berteh/musescore-chordsToNotes/)
//  - Jon Ensminger (AddNoteNameNoteHeads v. 1.2 plugin)
//  Thank you :-)
// 
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License version 2
//  as published by the Free Software Foundation
//=============================================================================
import MuseScore 3.0
import QtQuick 2.9
import QtQuick.Controls 2.2
import QtQuick.Layouts 1.2
import QtQuick.Dialogs 1.2
import QtQuick.Window 2.2
import Qt.labs.settings 1.0
MuseScore {
  property string pname:"Chord ID and Roman numeral analysis (dev)" //windowed diff find //DOCKED AND //REPEAT
  menuPath: "Plugins."+pname
  version: "20230827A"  //based on animikatamon's Chord Identifier (Pop & Jazz) version 3.4
  id: pluginscope
  //pluginType: "dock"; implicitHeight: 9999   //DOCKED
  Component.onCompleted:{
    if (mscoreMajorVersion==3){
      // pluginscope.pluginType="dock"
      // pluginscope.anchors.fill=Qt.binding(function(){return pluginscope.parent})  //responsive
    }
    if (mscoreMajorVersion >= 4) {
      // pluginscope.pluginType="dialog"
      pluginscope.title = pluginscope.pname
      // some_id.thumbnailName = "thumbnail.png";
      // some_id.categoryCode = "some_caregory";
    }
    if (  mscoreMajorVersion < 3  || (mscoreMajorVersion == 3 && mscoreMinorVersion < 3) ){
      console.log('This plugin requires MuseScore 3.3 and above')
      return
    }
    initdata()
    //initandstyle()  //DOCKED
  }
  Settings {
    id: settings
    category: "cid4"
    property int existingsym          : 0
    property int entire_note_duration : 1
    property int symformat            : 6  // coupled to  halfmodel's array
    property int sympos               : 3
    property var symColor             : "#000000"
    property int keyedit              : 0
    property int keysig               : 7
    property int keymM                : 1
    property int rnainvobject         : 2
    property int partialChords2       : 2
    property var partialChordsColor   : "#0000ff"
    property int showrepeat           : 0
    property int listnotes            : 0
    property int colortones           : 0
    property var colorNCT             : "#cc8800"    //non chord tone
    property var colorR               : "#000000"    //Root
    property var color3               : "#808000"    //3rd (4th for sus4)
    property var color5               : "#008000"
    property var color7               : "#804080"
    property var color9               : "#C080C0"
    property var color11              : "#80C0C0"
    property var color13              : "#4080C0"
    property int engedit              : 0
    // property int rom24                : 0
    // property int rom74                : 0
    property int jazz6                : 0
    property int jazzm6               : 0
    property int jazz69               : 0
    property int jazz24               : 0
    property int runandexit           : 1
  }
  property var romanStyle:[
    // old code, combined unicode ["Ⅰ","Ⅱ♭","Ⅱ","Ⅲ♭","Ⅲ","Ⅳ","Ⅴ♭","Ⅴ","Ⅵ♭","Ⅵ","Ⅶ♭","Ⅶ"]
    // [//jazzsym/major centric; wrong/unorthodox for minors analysis
    //    ['I','bII','II','bIII','III','IV','bV','V','bVI','VI','bVII','VII']
    //   ,['I','bII','II','bIII','III','IV','bV','V','bVI','VI','bVII','VII']
    // ],[//ABRSM  @msfp
       ['I','bII','II','bIII','III','IV','bV','V','bVI','VI','bVII','VII']
      ,['I','bII','II','III','hIII','IV','bV','V','VI','hVI','VII','VII'] //two VII: leading dim and subtonic/neapolitan/backdoor
    // ]
  ]
  property var chordStyles:[] //see initdata()
  function convertarray(xs,i,o){return xs.map(function(x){return x.replace(i,o)})}
  function invmodified(invmode,usechord,rv,rootpitch,keysig){
    var itv7symton=[//interval of the 7th from root based on current symbol tonality  minor7(dom7)=10, major7=11
       //I    II b  III IV    V  b VI    VII
       [11,10,10,11,10,11,10,10,11,10,10,10]
       // I7,IV7,V7 are chords of Major key I7,IV7=11, V7=10
       // bIII7,bVI7 are Major modal mixture of minor key =11
       // remaining 7ths are secondary dominants' =10
       //
       //i    ii   iii iv     v bvi      vii   minor only not affecting half or full diminshed chords
      ,[10,11,10,11,10,10,11,10,11,10,10,10]
       // ii7,iii7 are chords of Major key =10
       // i7,biv7 are chords of minor key i7,iv7=10
       // remaining 7ths are rarely used
       //   use notes of harmonic minor scale
       //   use flat bii,bv,bvi,bvii chord
    ]
    var t=pluginscope.chordStyles[usechord.type]
    var re=t[pluginscope.e.ROMNUMENABLED]
    var small=t[pluginscope.e.ROMNUMSMALL]
    //
    var has5mod7=t[pluginscope.e.ROMNUMINV]===4&&(
        ((re===40||re===60)&&itv7symton[small===1?1:0][getRomanSemi(rootpitch,keysig)]!==10)
      ||((re===41||re===61)&&itv7symton[small===1?1:0][getRomanSemi(rootpitch,keysig)]!==11))
    var invs=pluginscope.invStyles[invmode]
      [ t[rv?pluginscope.e.ROMNUMINV:pluginscope.e.INVTYPE]+(rv&&has5mod7?1:0) ]
    if(rv){
      //5
      invs=convertarray(invs,'f',
         (t[pluginscope.e.ROMNUMINV]===2||re===40||re===41)?'#'
        :(re===60||re===61)?'b'
        :''
      )
      //7
      var needb=((re===30||re===40||re===60)&&itv7symton[small===1?1:0][getRomanSemi(rootpitch,keysig)]!==10)
      var needs=((re===31||re===41||re===61)&&itv7symton[small===1?1:0][getRomanSemi(rootpitch,keysig)]!==11)
      if(!needb&&!needs){
        invs=convertarray(invs,'1s','')
          if(t[pluginscope.e.ROMNUMINV]===3){
          //fix 65 43 42   https://musescore.org/en/node/345409#comment-1204181
            invs=convertarray(invs,'6\n5s','65')
            invs=convertarray(invs,'4\n3s','43')
            invs=convertarray(invs,'4\n2\n','42')
          }
      }
      invs=convertarray(invs,'s', needb?'b' :needs?'#' :'' )
    }
    return invs
  }
  property var invStyles :[]
  property var e:[]
  function initdata(){
    pluginscope.e.JAZZENABLED = 0;
    pluginscope.e.QUALITY = 1;
    pluginscope.e.INTERVALS = 2;
                //value   0    1   2    3        4   5          6      7    8         9     10    11
                //numeric R,  b9,  9,  m3(#9),  M3, 11(sus4), #11(b5), 5,  #5(b13),  13(6),  7,   M7  //Ziya
    pluginscope.e.COLORENHAR = 3;
    pluginscope.e.INVTYPE = 4;  // invStyles
    pluginscope.e.ENHARTPC = 5
    pluginscope.e.ROMNUMENABLED = 6;
    pluginscope.e.ROMNUMQUA = 7;
    pluginscope.e.ROMNUMSMALL = 8;
    pluginscope.e.ROMNUMINV = 9;  // invStyles VERTNUM
    //
    pluginscope.chordStyles=[
    // 0  1            2              3    4  5    6  7   8  9     //the number in trailing comment is retained from a predecessor plugin for ref and is not used internally
      [1, ""          ,[4,7]         ,''  ,0 ,0   ,1 ,'' ,0 ,0   ],//00: M (0)
      [1, "m"         ,[3,7]         ,''  ,0 ,0   ,1 ,'' ,1 ,0   ],//01: m
      [1, "dim"       ,[3,6]         ,''  ,0 ,0   ,1 ,'o',1 ,0   ],//02: dim
      [5, "sus2"      ,[2,7]         ,''  ,6 ,0   ,0 ,'' ,0 ,0   ],//    sus2  = Suspended Two  Default disabled, recognized as 03: sus4 eg. c,d,g = Gsus4/C
      [6, "sus4"      ,[5,7]         ,''  ,0 ,0   ,0 ,'' ,0 ,0   ],//03: sus4 = Suspended Fourth
      [6, "7sus4"     ,[5,7,10]      ,''  ,1 ,0   ,0 ,'' ,0 ,0   ],//04: 7sus4 = Dominant7, Suspended Fourth
      [2, "6"         ,[4,7,     9]  ,'13',5 ,0   ,0 ,'' ,0 ,0   ],//    6  = Sixth  Default disabled, recognized as 07: m7  eg. c,e,g,a = Am7/C
      [1, "Maj7"      ,[4,7,11]      ,''  ,1 ,0   ,31,'' ,0 ,3   ],//05: M7 = Major Seventh
      [1, "m(Maj7)"   ,[3,7,11]      ,''  ,1 ,0   ,31,'' ,1 ,3   ],//06: mMa7 = minor Major Seventh
      [1, "7"         ,[4,7,10]      ,''  ,1 ,0   ,30,'' ,0 ,3   ],//08: 7 = Dominant Seventh   //switched priority with 07
      [1, "m7"        ,[3,7,10]      ,''  ,1 ,0   ,30,'' ,1 ,3   ],//07: m7 = minor Seventh
      [1, "o7"        ,[3,6,9]       ,''  ,1 ,[3] ,1 ,'o',1 ,1   ],//09: dim7 = Diminished Seventh
      [1, "Maj7(#5)"  ,[4,8,11]      ,'#5',1 ,0   ,41,'' ,0 ,4   ],//10: #5Maj7 = Major Seventh, Raised Fifth
      [1, "mMaj7(#5)" ,[3,8,11]      ,'#5',0 ,0   ,41,'' ,1 ,4   ],//
      [1, "7(#5)"     ,[4,8,10]      ,'#5',1 ,0   ,40,'' ,0 ,4   ],//11: #57 = Dominant Seventh, Raised Fifth
      [0, "m7(#5)"    ,[3,8,10]      ,'#5',1 ,0   ,0 ,'' ,1 ,4   ],//          all disabled 20230827 coz C D E G = I not iii7#5  ; chord = enhar. add9
      [1, "+"         ,[4,8]         ,'#5',0 ,[0] ,1 ,'' ,0 ,2   ],//12: #5 = Major Raised Fifth
      [3, "m6"        ,[3,7,     9]  ,'13',5 ,0   ,0 ,'' ,1 ,0   ],//    m6 = Minor Sixth  Default disabled, recognized as 13: m7b5 eg. c,e,g,a = Am7b5/C
      [1, "07"        ,[3,6,10]      ,''  ,1 ,0   ,1 ,'0',1 ,1   ],//13: m7b5 = minor 7th, Flat Fifth
      [1, "mMaj7(b5)" ,[3,6,11]      ,''  ,1 ,0   ,0 ,'' ,1 ,4   ],//          RNA disabled  20230827 coz C Db E G = I not bii7b5
      [1, "7(b5)"     ,[4,6,10]      ,''  ,1 ,0   ,60,'' ,0 ,4   ],//14: M7b5 = Major 7th, Flat Fifth
      [0, "Maj7(b5)"  ,[4,6,11]      ,''  ,1 ,0   ,61,'' ,0 ,4   ],//       RNA only; enhar. Maj7(#11)
      [1, "(add9)"    ,[4,7,   2]    ,''  ,2 ,0   ,0 ,'' ,0 ,0   ],//15: add9 = Major additional Ninth
      [1, "Maj9"      ,[4,7,11,2]    ,''  ,2 ,0   ,0 ,'' ,0 ,0   ],//16: Maj7(9) = Major Seventh, plus Ninth
      [1, "9"         ,[4,7,10,2]    ,''  ,2 ,0   ,0 ,'' ,0 ,0   ],//17: 7(9) = Dominant Seventh, plus Ninth
      [1, "m(add9)"   ,[3,7,   2]    ,''  ,2 ,0   ,0 ,'' ,1 ,0   ],//18: add9 = minor additional Ninth
      [1, "m9(Maj7)"  ,[3,7,11,2]    ,''  ,2 ,0   ,0 ,'' ,1 ,0   ],//19: m9(Maj7) = minor Major Seventh, plus Ninth
      [1, "m9"        ,[3,7,10,2]    ,''  ,2 ,0   ,0 ,'' ,1 ,0   ],//20: m7(9) = minor Seventh, plus Ninth
      [1, "Maj7(#11)" ,[4,7,11,  6]  ,'#4',3 ,0   ,0 ,'' ,0 ,0   ],//21: Maj7(#11) = Major Seventh, Sharp Eleventh
      [1, "Maj9(#11)" ,[4,7,11,2,6]  ,'#4',3 ,0   ,0 ,'' ,0 ,0   ],//22: Maj9(#11) = Major Seventh, Sharp Eleventh, plus Ninth 
      [1, "7(#11)"    ,[4,7,10,  6]  ,'#4',3 ,0   ,0 ,'' ,0 ,0   ],//23: 7(#11) =  Dom. Seventh, Sharp Eleventh
      [1, "9(#11)"    ,[4,7,10,2,6]  ,'#4',3 ,0   ,0 ,'' ,0 ,0   ],//24: 9(#11) =  Dom. Seventh, Sharp Eleventh, plus Ninth 
      [1, "7(13)"     ,[4,7,10,  9]  ,'13',4 ,0   ,0 ,'' ,0 ,0   ],//25: 7(13) =  Dom. Seventh, Thirteenth
      [1, "9(13)"     ,[4,7,10,2,9]  ,'13',4 ,0   ,0 ,'' ,0 ,0   ],//26: 9(13) =  Dom. Seventh, Thirteenth, plus Ninth 
      [1, "7(b9)"     ,[4,7,10,1]    ,''  ,2 ,0   ,0 ,'' ,0 ,0   ],//27: 7(b9) = Dominant Seventh, plus Flattened Ninth
      [1, "7(b13)"    ,[4,7,10,  8]  ,''  ,4 ,0   ,0 ,'' ,0 ,0   ],//28: 7(b13) =  Dom. Seventh, Flattened Thirteenth
      [1, "7(b9/b13)" ,[4,7,10,1,8]  ,''  ,4 ,0   ,0 ,'' ,0 ,0   ],//29: 7(b13b9) =  Dom. Seventh, Flattened Thirteenth, plus Flattened Ninth 
      [1, "7sus"      ,[4,7,10,    5],''  ,4 ,0   ,0 ,'' ,0 ,0   ],//    7(11)          @elsewhere
      [1, "7susb9"    ,[4,7,10,1,  5],''  ,4 ,0   ,0 ,'' ,0 ,0   ],//    7(b911)        @elsewhere
      [1, "11(b9/b13)",[4,7,10,1,8,5],'11',4 ,0   ,0 ,'' ,0 ,0   ],//30: 7(b13b911) =  Dom. Seventh, Flattened Thirteenth plus Flattenet Ninth, plus Eleventh 
      [1, "7(#9)"     ,[4,7,10,3]    ,'#9',2 ,0   ,0 ,'' ,0 ,0   ],//31: 7(#9) = Dominant Seventh, plus Sharp Ninth
      [4, "6/9"       ,[4,7,   2,9]  ,'13',2 ,0   ,0 ,'' ,0 ,0   ],//    6(9) = Nine Sixth Default disabled, recognized as 32: m7(11) 
      [1, "m7(11)"    ,[3,7,10,  5]  ,'11',3 ,0   ,0 ,'' ,1 ,0   ],//32: m7(11) = minor Seventh, plus Eleventh
      [1, "m11"       ,[3,7,10,2,5]  ,'11',3 ,0   ,0 ,'' ,1 ,0   ],//33: m9(11) = minor Seventh, plus Eleventh, plus Ninth 
      [1, ""          ,[  7]         ,''  ,0 ,0   ,0 ,'' ,0 ,0   ],// power chord    see "// 37 over powerchord"
    ]
    //
    pluginscope.e.HIDEINV=99 
    pluginscope.e.SUPER=0 //superscript inversion function has been disabled
    pluginscope.e.ABCD=1
    pluginscope.e.VERTNUM=2
    const superscript=['\u2070', '\u00B9', '\u00B2', '\u00B3', '\u2074', '\u2075', '\u2076', '\u2077', '\u2078', '\u2079']
    const subscript  =['\u2080', '\u2081', '\u2082', '\u2083', '\u2084', '\u2085', '\u2086', '\u2087', '\u2088', '\u2089']
    var clumpsyrna=function(r){return [0,1,2,3,4,5,6].map(function(s,i){return r+(i>0?'_inv'+s:'')})}
    pluginscope.invStyles=[ 
     [ //invmode==0 //superscript inversion function has been disabled
       [0,1,2]      .map(function(_,i){return i>0?' '+superscript[i]:'' })                                     //default INVTYPE==0
      ,[0,1,2,3,4]  .map(function(_,i){return i>0?' '+superscript[i]:'' })                                     //7       INVTYPE==1
      ,[0,1,2,3,4,5,6].map(function(_,i){return i>0?' '+superscript[9]                +'-'+superscript[i]:''}) //9       INVTYPE==2
      ,[0,1,2,3,4,5,6].map(function(_,i){return i>0?' '+superscript[1]+superscript[1] +'-'+superscript[i]:''}) //11      INVTYPE==3
      ,[0,1,2,3,4,5,6].map(function(_,i){return i>0?' '+superscript[1]+superscript[3] +'-'+superscript[i]:''}) //13      INVTYPE==4
      ,[0,1,2,3,4,5,6].map(function(_,i){return i>0?' '+superscript[6]                +'-'+superscript[i]:''}) //6       INVTYPE==5
      ,[0,1,2,3,4,5,6].map(function(_,i){return i>0?' '+superscript[2]                +'-'+superscript[i]:''}) //2       INVTYPE==6
    //    [0,1,2]      .map(function(_,i){return i>0?' <sup>'+i+'</sup>':'' })                   //default INVTYPE==0
    //   ,[0,1,2,3,4]  .map(function(_,i){return i>0?' <sup>'+i+'</sup>':'' })                   //7       INVTYPE==1
    //   ,[0,1,2,3,4,5,6].map(function(_,i){return i>0?' <sup>'+9 +'_'+i+'</sup>':''})           //9       INVTYPE==2
    //   ,[0,1,2,3,4,5,6].map(function(_,i){return i>0?' <sup>'+11+'_'+i+'</sup>':''})              //11      INVTYPE==3
    //   ,[0,1,2,3,4,5,6].map(function(_,i){return i>0?' <sup>'+13+'_'+i+'</sup>':''})              //13      INVTYPE==4
    //   ,[0,1,2,3,4,5,6].map(function(_,i){return i>0?' <sup>'+6 +'_'+i+'</sup>':''})             //6       INVTYPE==5
    //   ,[0,1,2,3,4,5,6].map(function(_,i){return i>0?' <sup>'+2 +'_'+i+'</sup>':''})              //2       INVTYPE==6
     ]
    ,[ //invmode ABCD
       '_'//default INVTYPE==0
      ,'_'//...
      ,'_'
      ,'_'
      ,'_'
      ,'_'
      ,'_'
     ].map(function(){return 'abcdddd'.split('').map(function(s){return ' '+s}) })
    ,[ //invmode VERTNUM  //0th item used in both (hideinversion) and (showinverions,isrootpos)
       ['','6','64']                                    // ROMNUMINV==0
      ,['7','65','43','42']                             // ROMNUMINV==1
      ,['+','6\n3f','+64']                              // ROMNUMINV==2
      ,['7s','6\n5s','4\n3s','4\n2\n1s']                // ROMNUMINV==3
      ,['7\n5f','6\n5\n3f','4\n3\n1f','6f\n4\n2']       // ROMNUMINV==4
      ,['7s\n5f','6\n5s\n3f','4\n3s\n1f','6f\n4\n2\n1s']// ROMNUMINV==4+1
     ]
    ]
    pluginscope.e.ROM=0
    pluginscope.e.JAZZ=1
  }
/**/
  function getNoteName(note_tpc){ 
    return ["Fbb","Cbb","Gbb","Dbb","Abb","Ebb","Bbb",
      "Fb","Cb","Gb","Db","Ab","Eb","Bb","F","C","G","D","A","E","B","F#","C#","G#","D#","A#","E#","B#",
      "F##","C##","G##","D##","A##","E##","B##"][note_tpc+1] }
  function getRomanSemi(rootpitch, keysig){
    return (rootpitch+(settings.keymM===0?3:0)+12-[11,6,1,8,3,10,5,0,7,2,9,4,11,6,1][keysig!=undefined?keysig:7])%12}
  function getRomanNum(rootpitch, keysig, small){
    var r=pluginscope.romanStyle[settings.keymM==0?1:0][getRomanSemi(rootpitch,keysig)]
    if(small===1) 'IV'.split('').map(function(n,i){r=r.replace(new RegExp(n,'g'),'iv'[i])})
    return r
  }
  function compare_arr(ref_arr, search_elt) {
    if (!ref_arr || !search_elt) return [];
    var its=[], fillcount=0;
    for(var i=0; i<ref_arr.length; i++){
      var match=~search_elt.indexOf(ref_arr[i]) 
      its[i]=match?ref_arr[i]:undefined
      if(match) fillcount++
    }
    return {
       its: its
      ,fillcount: fillcount
      ,fullfill: fillcount===ref_arr.length
    };
  }
  function analyseChord(notesset, romjazz) {
    var nsTones011 = notesset
      .map(function(x){return x.pitch%12})
      .sort(function(a,b){return a-b})
      .filter(function(v,i,a){return a.indexOf(v) === i})
    console.log('nsTones011 unique tones 0=C,  1=C# and Db etc : '+nsTones011)
    var len=nsTones011.length
    var nsTonesBassNudge=[]
    for(var i=0; i<len; i++){
      nsTonesBassNudge[i]=[]
      for(var k=0; k<len-1; k++){
        nsTonesBassNudge[i][k]=(24+ nsTones011[(i+k+1)%len] - nsTones011[i])%12
      }
    }
    console.log('nsTonesBassNudge semitones from the s tone (s = the n-th nsTones011)\n  , aka invertions of the notesset\n  later usechord.pos=s ; var rootTone011= nsTones011[usechord.pos]')
    console.log(nsTonesBassNudge.map(function(x,i){return '  [s='+i+']:'+JSON.stringify(x)}).join('  '))
    //
      var fillmatched=[], partialmatched=[]
      for(var t=0; t<pluginscope.chordStyles.length; t++){ //chord types  front priority
        console.log('testing chordStyles: [QUALITY] '+pluginscope.chordStyles[t][pluginscope.e.QUALITY]+', [INTERVALS] '+pluginscope.chordStyles[t][pluginscope.e.INTERVALS])
        if(
          (romjazz===pluginscope.e.ROM
            &&
            (  pluginscope.chordStyles[t][pluginscope.e.ROMNUMENABLED]!==0
            // ||(pluginscope.chordStyles[t][ROMNUMENABLED]===2 && settings.rom24==1)
            // ||(pluginscope.chordStyles[t][ROMNUMENABLED]===3 && settings.rom24==2)
            // ||(pluginscope.chordStyles[t][ROMNUMENABLED]===4 && settings.rom74==1)
            )
          )
          ||
          (romjazz===pluginscope.e.JAZZ
            &&
            (  pluginscope.chordStyles[t][pluginscope.e.JAZZENABLED]===1
            ||(pluginscope.chordStyles[t][pluginscope.e.JAZZENABLED]===2 && settings.jazz6==1)
            ||(pluginscope.chordStyles[t][pluginscope.e.JAZZENABLED]===3 && settings.jazzm6==2)
            ||(pluginscope.chordStyles[t][pluginscope.e.JAZZENABLED]===4 && settings.jazz69==1)
            ||(pluginscope.chordStyles[t][pluginscope.e.JAZZENABLED]===6 && (settings.jazz24==1 || settings.jazz24==2) )
            ||(pluginscope.chordStyles[t][pluginscope.e.JAZZENABLED]===5 && settings.jazz24==2)
            )
          )
        ){
          for(var s=0; s<nsTonesBassNudge.length; s++){
            var r = compare_arr(pluginscope.chordStyles[t][pluginscope.e.INTERVALS], nsTonesBassNudge[s])
            console.log('  s of nsTonesBassNudge: '+s)
            console.log('    r: '+JSON.stringify(r))
            if(r.fillcount>0){
              console.log('    at least one matched with [INTERVALS], either unshift() insert as 0-th or push() to fillmatched or partialmatched; pos=s')
              if(r.fullfill){
                fillmatched[
                  fillmatched.length>0 && fillmatched[0].fillcount<r.fillcount
                  ?'unshift':'push']({pos:s, type:t, fillcount:r.fillcount})
              }else{
                partialmatched[
                  partialmatched.length>0 && partialmatched[0].fillcount<r.fillcount
                  ?'unshift':'push']({pos:s, type:t ,its:r.its, fillcount:r.fillcount })
              }
              console.log('      fillmatched[]: '+fillmatched.map(function(x,i){return '['+i+']:'+JSON.stringify(x)}).join('  '))
              console.log('      partialmatched[]: '+partialmatched.map(function(x,i){return '['+i+']:'+JSON.stringify(x)}).join('  '))
            }else{console.log('  no match with [INTERVALS]')}
          }
        }else{console.log('  skipped, disabled by GUI')}
      }
      //
      var usechord
      if( partialmatched.length>0
          &&(    fillmatched.length==0
              || (fillmatched.length>0 && fillmatched[0].fillcount<partialmatched[0].fillcount) // 37 over powerchord
            )
      ){
        usechord=partialmatched[0]
        // match m3/M3 and 7/M7
        for(var i=0; i<partialmatched.length; i++){
          if( usechord.fillcount<=partialmatched[i].fillcount
              && ([3,4].filter(function(k){return ~partialmatched[i].its.indexOf(k)}).length>0)
              && ([10,11].filter(function(k){return ~partialmatched[i].its.indexOf(k)}).length>0)
          ){
            usechord=partialmatched[i]
            break;
          }
        }
        if(!usechord){
          // match m3/M3 
          for(var i=0; i<partialmatched.length; i++){
            if( ([3,4].filter(function(k){return ~partialmatched[i].its.indexOf(k)}).length>0)
            ){
              usechord=partialmatched[i]
              break;
            }
          }
        }
        console.log('rootTone011 = partialmatched[x].pos, coz best fit the src logic with this as root')
      }else if(fillmatched.length>0){
        usechord=fillmatched[0]
        console.log('rootTone011 = fillmatched[0].pos coz all [INTERVALS] are present with this as root, and is last inserted as 0-th')
      }
    if(usechord){
      var completeChord= fillmatched.length>0
      var rootTone011= nsTones011[usechord.pos]
      var c=pluginscope.chordStyles[usechord.type]
      var enhbytpc=c[pluginscope.e.ENHARTPC]
      var tones=c[pluginscope.e.INTERVALS].map(function(i){return (i+rootTone011)%12 })
      tones.unshift(rootTone011)
      console.log('tones[] = tone values in [rootTone011,rootTone011+...chordStyles[INTERVALS]]) : '+tones)
      //
      if(completeChord&&enhbytpc){
        console.log('override rootTone011, coz ENHARTPC = array (dim7 and aug), use ENHARTPC[0] value (y)')
        var notesbyTones=[]
        notesset.map(function(x){
          var i=tones.indexOf(x.pitch%12)
          if(~i){ //filter out NCT
            notesbyTones[i]=(notesbyTones[i]||[])
            notesbyTones[i].push(x)
          }
        })
        console.log('notesbyTones = *notesset* notes according to matching tone values in tones[]')
        console.log('  '+notesbyTones.map(function(x,i){return "[tones's i="+i+']:[{pitch: '+x[0].pitch+', tpc: '+x[0].tpc+'...}(,...)]'}).join('  '))
        rootTone011=
          notesbyTones
          .map(function(x){return x[0]}) //arbitrarily use one only
          .sort(function(a,b){return a.tpc-b.tpc})
          .filter(function(_,i){return i===enhbytpc[0]})[0].pitch%12
        console.log("sort notesbyTones by tpc, arbitrarily use first one of each")
        console.log("  rootTone011= the y-th sorted item: "+rootTone011)
      }
      return {
         usechord:      usechord
        ,rootTone011:   rootTone011
        ,completeChord: completeChord
      }
    }else{console.log('Not analysed notesset : '+notesset)}
  }
  function getRoot(notesset, rootTone011){
    var root
    notesset.map(function(y){
      if(y.pitch%12 === rootTone011) root=y
    })
    return root
  }                                   //TODO handle not found
  function getColors(notesset, adata){
    var chordcolor=[]
    if (adata.usechord) {
      for(var i=0; i<notesset.length; i++){
        var ch=pluginscope.chordStyles[adata.usechord.type][pluginscope.e.COLORENHAR]
        var p=notesset[i].pitch%12
        var testroot= p === adata.rootTone011
        var itv=function(k){return ~pluginscope.chordStyles[adata.usechord.type][pluginscope.e.INTERVALS].indexOf(k) && p === (adata.rootTone011+k)%12 }
        var test3 =                      itv(4)||                            (itv(3)&&ch!=='#9')||(itv(5)&&ch!=='11')
        var test5 = (itv(6)&&ch!=='#4')||itv(7)        ||(itv(8)&&ch==='#5')
        var test7 =                      itv(10)||itv(11)                                                         ||(itv(9)&&ch!=='13')
        var test9 =                      itv(1)||itv(2)||                    (itv(3)&&ch==='#9')
        var test11= (itv(6)&&ch==='#4')                                                         ||(itv(5)&&ch==='11')
        var test13=                                      (itv(8)&&ch!=='#5')                                      ||(itv(9)&&ch==='13')
        // notesset[i].color=
        chordcolor.push({pitch: notesset[i].pitch, tpc: notesset[i].tpc ,color:
            testroot?settings.colorR 
          : test3?settings.color3
          : test5?settings.color5
          : test7?settings.color7
          : test9?settings.color9
          : test11?settings.color11
          : test13?settings.color13
          : settings.colorNCT
        })
      }
    }
    return chordcolor
  }
  function getInversion (notesset, adata, rootTone011){
    var inversion
    var tones=pluginscope.chordStyles[adata.usechord.type][pluginscope.e.INTERVALS].map(function(i){return (i+rootTone011)%12 })
    tones.unshift(rootTone011)
    notesset.map(function(y){
      if(inversion===undefined){
        var m=tones.indexOf(y.pitch%12)
        if(~m) inversion=m
      }
    })
    return (inversion||0)
  }
  function getNametext(notesset, symformat, adata, root, inversion){
    var invmode=
       symformat===1? pluginscope.e.VERTNUM
      :symformat===2? pluginscope.e.ABCD
      :symformat===3? pluginscope.e.VERTNUM
      :symformat===4? pluginscope.e.SUPER
      :pluginscope.e.HIDEINV
    var text=[]
    text[0] = 1<=symformat&&symformat<=3?(
      getRomanNum(root.pitch, settings.keysig, pluginscope.chordStyles[adata.usechord.type][pluginscope.e.ROMNUMSMALL])
      + pluginscope.chordStyles[adata.usechord.type][pluginscope.e.ROMNUMQUA]
    ):(
      getNoteName(root.tpc)
      + pluginscope.chordStyles[adata.usechord.type][pluginscope.e.QUALITY]
    )
    text[1] = 1<=symformat&&symformat<=3?(
      invmodified( invmode ,adata.usechord ,symformat===1||symformat===3 ,root.pitch , settings.keysig )
      [symformat===3?0:inversion]
    ):(symformat===5 && inversion>=1)?(
      "/"+getNoteName(notesset[0].tpc)
    ):''
    return text
  }
  function chordDuration(chord){
    var duration = chord.globalDuration;    // only from MS 3.5 onwards
    if ( !duration ) duration = chord.duration
    return duration.ticks
  }
  function getAllCurrentNotes(cursor, startStaff, endStaff, onlySelected, prev_chord){
    var cur = [], pedals =[]
    for (var staff = endStaff; staff >= startStaff; staff--) {
      for (var voice = 3; voice >=0; voice--) {
        cursor.track = (staff*4)+voice;
        if (cursor.element && cursor.element.type == Element.CHORD) {
          var notes = cursor.element.notes;
          for (var i = 0; i < notes.length; i++) {
            if (onlySelected && !notes[i].selected)
              continue;
            cur.push(notes[i]);
          }
        }
      }
    }
    if (prev_chord) {
      for (var i = 0; i < prev_chord.length; i++) {
        var note = prev_chord[i];
        if(
          ( cursor.tick < note.parent.parent.tick + chordDuration(note.parent)  )
          && 1===settings.entire_note_duration
        ) pedals.push(note);
      }
    }
    return {cur:cur,pedals:pedals};
  }
  function closestNextElement(cursor, startStaff, endStaff, elemType) {
    var seg = cursor.segment;
    if ( !seg ) return false;
    while(seg = seg.next) {
      var tr;
      for (tr = 0; tr<=endStaff*4+3; tr++) {
        var el = seg.elementAt(tr);
        if (el && el.type == elemType){
          cursor.track = tr;
          while (cursor.tick < seg.tick)  cursor.next();
          if (cursor.tick > seg.tick)  console.log('Error at closestNextElement()');
          return true;
          break
        }
        if(tr==0 && startStaff!=0) tr=startStaff*4-1
      }
    }
     return false
  }

  property var timeout
  Component { id: setTimeout ; Timer {  } }
  function runsheet() {
    //
      function copycursor(cursor){
        var nc=curScore.newCursor()
        nc.track=cursor.track
        nc.rewindToTick(cursor.tick)
        return nc
      }
      function addToScore(e,cursor,track){ //add to top but need hacking for non occupied tick at track0
        var c=curScore.newCursor()
        c.track=cursor.track //ms3 bug ?hacky
        c.rewindToTick(cursor.tick)
        c.track=track //ms3 bug ?hacky
        c.add(e)
      }
      var map=function(o,f){var x=[];for(var i in o){x.push(f(o[i],i))};return x}
      var filter=function(o,f){var x=[];for(var i in o){if(f(o[i],i)==true) x.push(o[i])};return x}
      //
      var selnotes=filter(curScore.selection.elements,function(e){return e.type===Element.NOTE})
      var uniqueandsort=function(arr,func){
        return arr.map(func)
        .filter(function(v,i,a){return a.indexOf(v) === i})
        .sort(function(a,b){return a-b})
      }
      var selOnTicks=uniqueandsort(selnotes, function(e){return e.parent.parent.tick} )
      var selOffTicks=uniqueandsort(selnotes, function(e){return e.parent.parent.tick+chordDuration(e.parent)})
      var selInstrumentTopStaffv1=uniqueandsort(selnotes, function(e){return e.staff.part.startTrack})
      var selInstrumentLastStaffv1=uniqueandsort(selnotes, function(e){return e.staff.part.endTrack-3 /*ms3 bug*/-1 })
      var selStaffs=uniqueandsort(selnotes, function(e){return Math.floor(e.track/4)})
    //
      var cursor = curScore.newCursor()
        ,startStaff = 0
        ,endStaff = curScore.nstaves - 1
        ,fullScore = true
        ,starttick = 0   // cursor.rewind(Cursor.SCORE_START);  // start from beginning of score
        ,endtick = curScore.lastSegment.tick + 1
        ,ist = 0
        ,iet = curScore.ntracks-3 /*ms3 bug*/-1 
      if(selnotes.length>0){
        fullScore=false
        startStaff=selStaffs[0]
        endStaff=selStaffs.slice(-1)[0]
        starttick=selOnTicks[0]
        endtick=selOffTicks.slice(-1)[0]
        ist = selInstrumentTopStaffv1[0]
        iet = selInstrumentLastStaffv1.slice(-1)[0]
      }
      cursor.track=startStaff*4
      cursor.rewindToTick(starttick)
      // var keySig = cursor.keySignature; //ms3 bug only 1st keysig at 1st measure of 1st staff
    //
      var romjazz=
        0===settings.symformat?undefined
        :1<=settings.symformat&&settings.symformat<=3?pluginscope.e.ROM
        :pluginscope.e.JAZZ
      //
      var sympostrack=
         settings.sympos===0? curScore.ntracks-3 /*ms3 bug*/-1 
        :settings.sympos===1? iet
        :settings.sympos===2? ist
        :settings.sympos===3? 0
        : 0 //ms3 bug ?hacky
    function workaseg(cursor, prevseg){
      const AC = 1
      const ECS = 2
      pbar.value = (cursor.tick-starttick)/(endtick-starttick)
      var segment
      if (segment=cursor.segment){
        var propagate={} //next workaseg's prevseg
        var readonlycursor=copycursor(cursor) //readonly til end of IF bracket, closestNextElement() contains cursor.next() that mutates it
        //
        var gacn = getAllCurrentNotes(cursor, startStaff, endStaff, !fullScore, (prevseg.notesset||[]) );
        var notesset=[].concat(gacn.cur).concat(gacn.pedals)
        if(notesset.length>0){
          var usingExistingText
          var reuse={}
          filter(segment.annotations,function(){return true})
          .map(function(e){
            if(e.type===Element.HARMONY && settings.existingsym===2){
              removeElement(e)
            }else if(e.type===Element.HARMONY && settings.existingsym===1){
              e.text='('+e.text+')'
            }else if(settings.existingsym===3){
              if(e.type===Element.HARMONY ||e.type===Element.ROMAN || e.text=='-'){
                usingExistingText=true  //suppress chord id with any text 
                if( e.type===Element.HARMONY 
                  && e.text
                    .replace('dim','')
                    .replace('aj','')
                    .replace('b5','')
                    .replace(/\/.*/,'')
                    .match(/[A-Ga-g]/)
                ){
                  reuse.chordText=romjazz===pluginscope.e.JAZZ?[e.text,'']:undefined              //  jazz ecs-> RNA
                    reuse.chordNorm=e.text.replace(/\/.*/,'').toLowerCase()
                    reuse.inversion=0
                    reuse.colorProg=ECS
                }else if(e.text=='-'){
                  if(prevseg.chordNorm){
                    reuse.chordText=prevseg.chordText
                    reuse.chordNorm=prevseg.chordNorm
                    reuse.inversion=prevseg.inversion
                    reuse.colorProg=prevseg.colorProg
                  }
                }
              }
            }
          })
          var adata
          var aInversion
          var aChordText
          var aCompleteChord
          var aChordNorm
          var writeSym={}
          var lncheck
          notesset.sort(function(a,b){return a.pitch-b.pitch}) //bass = [0]
          if(notesset.length>0 && /*experimental skip*/!usingExistingText){
            adata=analyseChord(notesset, romjazz)
            if(adata){
              var root = getRoot(notesset, adata.rootTone011)
              aInversion = getInversion(notesset, adata, adata.rootTone011)
              aChordText = getNametext(notesset, settings.symformat, adata, root, aInversion) //array
              aCompleteChord = adata.completeChord
              aChordNorm = getNametext(notesset, 6/*settings.symformat*/, adata, root, 0).join('').toLowerCase() // noramlize to root pos = no position info
            }else{ //analyseChord() no result coz tones/mod 12 length===1, ie unison, 8ve etc , handle using logic similar to having an existing text '-'
                if(prevseg.chordNorm){
                  usingExistingText=true  //suppress chord id
                  reuse.chordText=prevseg.chordText
                  reuse.chordNorm=prevseg.chordNorm
                  reuse.inversion=prevseg.inversion
                  reuse.colorProg=prevseg.colorProg
                }//else{//anacrusis}
            }
          }
          if(usingExistingText||adata){
            propagate=
                                         usingExistingText&&  reuse.chordNorm                      ? {chordNorm:reuse.chordNorm, inversion:reuse.inversion, chordText:reuse.chordText, completeChord:true}
            : settings.symformat===0 || (usingExistingText&&  !reuse.chordNorm  )                  ? {}
            : aCompleteChord                                                                       ? {chordNorm:aChordNorm,      inversion:aInversion,      chordText:aChordText,      completeChord:aCompleteChord}
            : settings.partialChords2==0 || settings.partialChords2==1                             ? {}
            : settings.partialChords2==2 && aChordNorm==prevseg.chordNorm && prevseg.completeChord ? {chordNorm:aChordNorm,      inversion:aInversion,      chordText:aChordText,      completeChord:true}
            :                                                                                        {chordNorm:aChordNorm,      inversion:aInversion,      chordText:aChordText,      completeChord:aCompleteChord}
            propagate.notesset = notesset
            writeSym =
              settings.symformat===0 || usingExistingText || (!aCompleteChord && settings.partialChords2==0)                    ? {}
            :                                                 !aCompleteChord && settings.partialChords2==1                     ? {text:['X','']}
            :                                                 !aCompleteChord && settings.partialChords2==2                     ? {text:propagate.chordText}
            : settings.showrepeat===2                                                                                           ? {text:propagate.chordText}
            : settings.showrepeat===1&&prevseg.completeChord
              &&(aChordNorm==prevseg.chordNorm&&(settings.symformat===3||settings.symformat===6||aInversion==prevseg.inversion))? {text:['-','']}
            :    aChordNorm==prevseg.chordNorm&&aInversion==prevseg.inversion                                                   ? {}
            : aChordNorm!=prevseg.chordNorm  || aInversion!=prevseg.inversion                                                   ? {text:propagate.chordText}
            : /*settings.showrepeat===0*/                                                                                       {}
            propagate.colorProg = 
                                                                     usingExistingText&&  reuse.chordNorm     ? reuse.colorProg
            : settings.symformat===0  || settings.colortones===0 || (usingExistingText&&  !reuse.chordNorm)
              ||(!propagate.completeChord && (settings.partialChords2==0 || settings.partialChords2==1))      ? undefined
            :                                                                                                   AC
            propagate.adata=
                                                                     usingExistingText&&  reuse.chordNorm     ? prevseg.adata
            : settings.symformat===0  || settings.colortones===0 || (usingExistingText&&  !reuse.chordNorm)
              ||(!propagate.completeChord && (settings.partialChords2==0 || settings.partialChords2==1))      ? undefined
            :                                                                                                   adata
            lncheck = settings.symformat>0 && (  settings.listnotes==2 || (settings.listnotes==1 && !propagate.completeChord)  )
          }
          if(writeSym.text){
            var fb= romjazz===pluginscope.e.ROM&&settings.rnainvobject!==2 &&~writeSym.text[1].indexOf('\n')
            if(fb){
              if(settings.rnainvobject===0){
                var sym = newElement(Element.FIGURED_BASS)//Element.STAFF_TEXT)
                sym.text = writeSym.text[1]
                sym.color = !propagate.completeChord ? settings.partialChordsColor : settings.symColor;
                sym.placement = settings.sympos===2||settings.sympos===3? Placement.ABOVE : Placement.BELOW
                addToScore(sym,readonlycursor,sympostrack)
              }else if(settings.rnainvobject===1){
                var sym = newElement(Element.STAFF_TEXT)
                sym.text = writeSym.text[1]
                sym.color = !propagate.completeChord ? settings.partialChordsColor : settings.symColor;
                sym.placement = settings.sympos===2||settings.sympos===3? Placement.ABOVE : Placement.BELOW
                sym.fontSize=7
                sym.offsetX = 1
                sym.offsetY = 2.5
                addToScore(sym,readonlycursor,sympostrack)
              }
            }
            var sym = newElement(Element.HARMONY)
            sym.harmonyType = romjazz===pluginscope.e.ROM ? HarmonyType.ROMAN : HarmonyType.STANDARD
            sym.text = writeSym.text[0].replace('07',romjazz===pluginscope.e.JAZZ&&settings.jazzm6===1?'m7b5':'07')+(fb?'':writeSym.text[1])
            sym.color = !propagate.completeChord ? settings.partialChordsColor : settings.symColor;
            sym.play = false
            sym.placement = settings.sympos===2||settings.sympos===3? Placement.ABOVE : Placement.BELOW
            // sym.subStyle = romjazz===pluginscope.e.ROM ? Tid.HARMONY_ROMAN : Tid.HARMONY_A
            // sym.fontFace = romjazz===pluginscope.e.ROM ? 'Campania' : 'Edwin'
            addToScore(sym,readonlycursor,sympostrack)
          }
          if(propagate.colorProg===AC && propagate.adata){
            var aColor=getColors(notesset, propagate.adata)
            gacn.cur.map(function(e,i){
              aColor.map(function(c){
                if(c.pitch===e.pitch&&c.tpc===e.tpc) e.color=c.color })
            })
          }
          if(propagate.colorProg===ECS && propagate.chordNorm){
            var ct=chord2tone(propagate.chordNorm)
            var ct12=ct.tones.map(function(c){return c%12})
            var ctitv=ct.tones.map(function(c){return (108+c-ct.tones[0])%12 })  //108 A0 A9
            var cm=ct.chordMap
            var ecsColor=[0,1,2,3,4,5,6,7,8,9,10,11].map(function(pa){
              var k=ct12.indexOf(pa)
              var testroot= k===0
              var test3 =
                ctitv[k]===4
                ||(ctitv[k]===3&&cm[3]===-1)
                ||(ctitv[k]===5&&cm[4]===0)
              var test5 =
                (ctitv[k]===6&&cm[5]===-1)
                ||ctitv[k]===7
                ||(ctitv[k]===8&&cm[5]===1)
              var test7 =
                ctitv[k]===10||ctitv[k]===11
                ||(ctitv[k]===9&&cm[7]===-2)
              var test9 = 
                ctitv[k]===1||ctitv[k]===2
                ||(ctitv[k]===3&&cm[9]===1)
              var test11=
                (ctitv[k]===6&&cm[4]===1)
                ||(ctitv[k]===5&&cm[11]===0)
              var test13=
                (ctitv[k]===8&&cm[13]===-1)
                ||(ctitv[k]===9&&cm[13]===0)
              var color=
                  testroot?settings.colorR 
                : test3?settings.color3
                : test5?settings.color5
                : test7?settings.color7
                : test9?settings.color9
                : test11?settings.color11
                : test13?settings.color13
                : settings.colorNCT
              return {pitch:pa, color:color}
            })
            //
            gacn.cur.map(function(e,i){
              ecsColor.map(function(c){
              if(c.pitch===e.pitch%12) e.color=c.color })
            })
          }
          if(lncheck){
            var systextele = newElement(Element.STAFF_TEXT)
            systextele.text = '- '+notesset
              .map(function(n){return getNoteName(n.tpc)})
              .filter(function(v,i,a){return a.indexOf(v) === i})
              .reverse().join('\n- ')
            systextele.fontSize = 8
            systextele.fontStyle = 2
            systextele.placement = settings.sympos===2||settings.sympos===3? Placement.ABOVE : Placement.BELOW
            addToScore(systextele,readonlycursor,sympostrack)
          }
        }
        //closestNextElement mutates readonlycursor
        if ( closestNextElement(readonlycursor, startStaff, endStaff, Element.CHORD) &&  readonlycursor.tick < endtick) 
          setTimeout.createObject(pluginscope,{ interval:1 ,running:true  }).triggered.connect(function(){ 
            if(working) workaseg(readonlycursor, propagate) })
        else
          setTimeout.createObject(pluginscope,{ interval:1 ,running:true  }).triggered.connect(function(){ 
            pbar.value =0; finapp(settings.runandexit===1) })
      }
    }
    workaseg(cursor,{})
  }
/**/  // 2023 msfp adapated from ExpandChordSymbols.js of Expand Chord Symbols v4.0
  // mod extras: slash inside parentheses
  function parseChordSymbol2(symbol) {
      var fullSymbol = symbol;

      // If the entire symbol is enclosed in parentheses, we simply remove the parentheses.
      var tokens = symbol.match(/^\((.*)\)$/);
      if (tokens) {
          symbol = tokens[1];
      }

      // If the chord symbol contains 1 or more substrings enclosed in parentheses,
      // remove those strings from the symbol and save them in the "extras" array.
      // We will deal with these in a moment.
      var extras = [];
      while (true) {
          tokens = symbol.match(/\((.*?)\)/);
          if (!tokens) break;
          extras.push(tokens[1].replace(/\//g, ""));
          symbol = symbol.replace(/\((.*?)\)/, "");
      }

      var result = {sus: [], add: [], drop: [], alter: []};

      // Do the front stuff
      var regex = RegExp(frontStuff(), "g");
      tokens = regex.exec(symbol);
      if (tokens) {
          if (tokens[1]) result.letter = tokens[1];
          if (tokens[2]) result.sharp = true;
          if (tokens[3]) result.flat = true;
          if (tokens[4]) result.major = true;
          if (tokens[5]) result.minor = true;
          if (tokens[6]) result.diminished = true;
          if (tokens[7]) result.halfdim = true;
          if (tokens[8]) result.augmented = true;
          if (tokens[9]) result.triangle = true;
          if (tokens[10]) doNum(result, tokens[10], false);

          // trim the front stuff off the symbol
          symbol = symbol.substring(regex.lastIndex);
      }

      // do the middle stuff
      var temp = parseMiddleStuff(result, symbol);
      result = temp.result;
      symbol = temp.symbol;

      // do the back stuff
      if (symbol.length > 0) {
          temp = parseBackStuff(result, symbol);
          result = temp.result;
          symbol = temp.symbol;
          if (symbol.length > 0) {
              console.log("warning: we didn't fully parse the chord symbol", fullSymbol);
          }
      }

      // do the extras. Each extra contains "middle stuff".
      for (var i = 0; i < extras.length; i += 1) {
          temp = parseMiddleStuff(result, extras[i]);
          result = temp.result;
          if (temp.symbol.length !== 0) {
              console.log("warning: we didn't fully parse the chord symbol", fullSymbol);
          }
      }        

      // Uncomment the following line to see what the chord specification looks like.
      // console.log(symbol, JSON.stringify(result));

      return result;
  }
  function chord2tone(text) {
      var chordSpec = parseChordSymbol2(text);
      var chordMap = expandChordSpec(chordSpec);
      var tones=[]
      var notesemitones = {1: 0, 2: 2, 3: 4, 4: 5, 5: 7, 6: 9, 7: 11
      , 8: 0, 9: 2, 10: 4, 11: 5, 12: 7, 13: 9};
      if(chordSpec.letter) for (var i in chordMap) {
          tones.push((
            letterToMidiNote(chordSpec) 
            + notesemitones[i] 
            + chordMap[i]
          )%12)
      }
      return {tones:tones,chordMap:chordMap}
  }
/**/ // MIT ExpandChordSymbols.js https://github.com/markshepherd/ExpandChordSymbols 
  // ExpandChordSymbols.js
  // Copyright (c) 2020 Mark Shepherd
  // 
  // Permission is hereby granted, free of charge, to any person obtaining a copy
  // of this software and associated documentation files (the "Software"), to deal
  // in the Software without restriction, including without limitation the rights
  // to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  // copies of the Software, and to permit persons to whom the Software is
  // furnished to do so, subject to the following conditions:
  // 
  // The above copyright notice and this permission notice shall be included in all
  // copies or substantial portions of the Software.
  // 
  // THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  // IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  // FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  // AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  // LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  // OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  // SOFTWARE.

  // This plugin for MuseScore 3 generates notes for chord symbols. For each chord symbol in the 
  // score, the plugin creates a corresponding set of notes in a designated target staff. Each
  // chord plays until the next chord. There are 2 modes:
  // - raw mode: all the notes of the chord are generated
  // - condensed mode: we generate only the 4 most important notes plus a bass note,
  //   and is inverted so that most of the notes are near or below middle C.
  //   The results are almost identical to Marc Sabatella's proposed voicings at
  //   https://musescore.com/marcsabatella/chord-symbol-voicings-for-playback

  // In the following code, we use various data structures to represent notes and chords.
  //   string - a simple string
  //      e.g. "Cmin7"
  //   score chord symbol array - an array with one element for each chord symbol that appears in the score.
  //      e.g. [{tick: 1440, text: "Cmin7", duration: 480}, ...]
  //   chord specification - a detailed specification of a single chord
  //      e.g. {letter: "C", number: "7", minor: true}
  //   chordMap - an object that represents all the notes in a chord
  //      e.g. {"1": 0, "3": -1, "5": 0, "7": -1} for Cmin7 = tonic, flat 3rd, perfect 5th, flat 7th.
  //   interval - the number of semitones between two notes.
  //      e.g. the interval between C and D is 2, the interval between tonic and perfect fifth is 7
  //   midi note - a number that describes a certain pitch. Arrays of midiNotes should always be sorted ascending.
  //      e.g. [48, 51, 55, 58] = C3, Eb3, G3, Bb3 = Cmin7

  // To generate the chords, we start with the text chord symbols that appear in the score,
  // and perform a series of transformations.
  // 0. findAllChordSymbols() searches all tracks of the score to produce a score chord symbol array. 
  // 1. parseChordSymbol() parse the text chord symbol, and produces a chord specification.
  // 2. expandChordSpec() uses the chord specification to create an chordMap object.
  // 3. in condensed mode, prune() reduces the number of notes in the chordMap to maximum of 4.
  // 4. render() converts the chordMap to an array of real midi notes
  // 5. in condensed mode, findOptimumInversion() modifies the list of midi notes to get the optimum voicing
  //      e.g. [55, 58, 60, 63] = Cmin7 (notes 48 and 51 got bumped up an octave)
  // 6. [addBass] add the bass note to the chord.
  //      e.g. [48, 55, 58, 60, 63]

  // Returns the interval above C that corresponds to the letter/sharp/flat fields in "spec"
  function letterToInterval(spec) {
      var letterToSemi = {C: 0, D: 2, E: 4, F: 5, G: 7, A: 9, B: 11};
      var result = letterToSemi[spec.letter.toUpperCase()];
      if (spec.flat) result--;
      if (spec.sharp) result++;
      if (result > 11) result -= 12;
      if (result < 0) result += 12;
      return result;
  }

  // Returns a copy of the source object.
  function shallowCopy(source) {
      var result = {};
      for (var i in source) {
          result[i] = source[i];
      }
      return result;
  }

  // Returns a midiNote in the octave below middle C that corresponds to the letter/sharp/flat fields in "spec".
  function letterToMidiNote(spec) {
      return letterToInterval(spec) + 48;
  }

  // Returns the number of properties in an object. E.g. for {a: 1, b: true} we return 2.
  function countProperties(obj) {
      var result = 0;
      for (var i in obj) result += 1;
      return result;
  }

  // Given an array of midi notes, return the number of notes that are above the threshold
  function numNotesAboveThreshold(midiNotes, threshold) {
      var result = 0;
      for(var i in midiNotes) {
          if (midiNotes[i] > threshold) result += 1; // use c#, not c
      }
      return result;
  }

  // Adds a rest to the score, using the current cursor position and duration.
  function addRest(cursor) {
      // Adding a rest to the score requires a little dance ...

      // ... first we add a placeholder note.
      cursor.addNote(0);

      // ... go back to the note we just added
      cursor.prev();

      // ... create a new rest with the same duration as the placeholder object
      var e = newElement(Element.REST);
      e.durationType = cursor.element.durationType;
      e.duration = cursor.element.duration;

      // ... add the rest to the score
      cursor.add(e);

      // ... advance the cursor, because cursor.add() doesn't. (unlike cursor.addNote(), which does).
      cursor.next();
  }

  // Removes all the notes in the given track.
  function clearStaff(track) {
      var cursor = curScore.newCursor();
      cursor.track = track;
      cursor.rewind(Cursor.SCORE_START);
      while (cursor.segment) {
          if (cursor.element.type === Element.CHORD) {
              removeElement(cursor.element);
          }
          cursor.next();
      }
  }

  // Returns a list of all the notes in the current score's selection,
  // in the form [{duration: <num>, tick: <num>, rest: <bool>}, ...]
  // "tick" is relative to the start of the selection.
  function getSelectedRhythm() {
      var result = [];

      // locate the selection
      var cursor = curScore.newCursor();
      cursor.rewind(Cursor.SELECTION_START);

      if (cursor.segment) {
          // The selection exists. Remember where it begins.
          var startTick = cursor.tick;

          // We will use this variable to consolidate tied notes into a single note
          var firstNote;

          // We will only look at the first staff of the selection.
          var staffId = cursor.staffIdx;

          // Find the end of the selection
          cursor.rewind(Cursor.SELECTION_END);
          var endTick = (cursor.tick !== 0)
              ? cursor.tick                    // Normal case
              : curScore.lastSegment.tick + 1; // The selection includes the end of the last measure.

          // Now go back to the beginning of the selection, and iterate over all the notes.
          // The rewind() function sets the voice to 0, that's the only voice we will look at.
          cursor.rewind(Cursor.SELECTION_START); 
          cursor.staffIdx = staffId;
          while (cursor.segment && cursor.tick < endTick) {
              if (cursor.element) {
                  var duration = cursor.element.duration;
                  var durationInTicks = (division * 4) * duration.numerator / duration.denominator;
                  var resultNote = {duration: durationInTicks, tick: cursor.tick - startTick};

                  if (cursor.element.type === Element.CHORD) {
                      // TODO: if the note is part of a triplet, adjust the numbers as required.

                      // See if the current note is tied to the previous note, or the next note.
                      if (cursor.element.notes && cursor.element.notes.length > 0) {
                          var note = cursor.element.notes[0];

                          if (note.tieForward && !note.tieBack) {
                              // This is the first note of a tied sequence
                              firstNote = resultNote;
                          }
                          if (note.tieBack && firstNote) {
                              // This is a note of a tied sequence
                              firstNote.duration += resultNote.duration;
                              resultNote = null;
                          }
                          if (!note.tieForward) {
                              // This note is not tied to the next note.
                              firstNote = null;
                          }
                      }
                      if (resultNote) result.push (resultNote);
                  } else if (cursor.element.type == Element.REST) {
                      resultNote.rest = true;
                      result.push (resultNote);
                  }
              }
              cursor.next();
          }
      }

      return result.length > 0 ? result : null;
  }

  // A regular expression that matches the "front" part of a chord symbol. These items -
  // the chord root, the triad type, and the number - can only occur once.
  function frontStuff() {
      return ["^",
          "([A-Ga-g])",
          "([#♯])?",
          "([b♭])?",
          "(?:",
              "(Major|major|Maj|maj|Ma|ma|M|j)|",
              "(minor|min|mi|m|-|−)|",
              "(dim|o|°)|",
              "(ø|O|0)|",
              "(aug|\\+)|",
              "([tΔ∆\\^])",
          ")?",
          "([0-9]+)?"
      ].join("");
  }

  // A regular expression that matches the "middle" part of a chord symbol. These items -
  // added, suspended, dropped, and altered notes - can occur multiple times within a symbol.
  function middleStuff() {
      return [
          "(?:",
              "(69|6-9|6\\+9|6\\/9)|",
              "(?:(?:Major|major|Maj|maj|Ma|ma|M|j)([0-9]+))|",
              "(alt)|",
              "(sus([0-9])?)|",
              "(?:add([0-9]+))|",
              "(?:(?:drop|no)([0-9]+))|",
              "([0-9]+)|",
              "(?:([#♯])?([b♭])?([0-9]+))|",
          ")"
      ].join("");
  }

  // A regular expression that matches the "back" part of a chord symbol.
  // This is only the optional bass note.
  function backStuff() {
      return "(?:\\/([A-G])([#♯])?([b♭])?)";
  }

  // Determine the exact meaning of a number in a chord symbol, which depends on context. 
  function doNum(result, token, inMiddle) {
      if (token === "69") {
          result.sixnine = true;
      } else if (inMiddle) {
          result.add.push(token);
      } else {
          result.number = token;
      }
  }

  // Parse the middle section of a chord symbol, and add the information into "result". 
  // "symbol" is the original chord symbol with the front stuff removed.
  function parseMiddleStuff(result, symbol) {
      var regex = RegExp(middleStuff(),'g');
      var charsConsumed;
      while (true) {
          // regex.exec will find the next matching item.
          var tokens = regex.exec(symbol);

          // if nothing found, then we're done
          if (!tokens || !tokens[0] || tokens.index >= symbol.length || regex.lastIndex === 0) break;

          // Examine the match results and update result accordingly.
          charsConsumed = regex.lastIndex;
          if (tokens[1]) result.sixnine = true;
          if (tokens[2]) result.majoralt = tokens[2];
          if (tokens[3]) result.alt = true;
          if (tokens[4]) result.sus.push(tokens[5] || "4");
          if (tokens[6]) result.add.push(tokens[6]);
          if (tokens[7]) result.drop.push(tokens[7]);
          if (tokens[8]) doNum(result, tokens[8], true);
          if (tokens[11]) result.alter.push({sharp: tokens[9], flat: tokens[10], number: tokens[11]});
      }

      return {result: result, symbol: symbol.substring(charsConsumed)};
  }

  // Parse the back section of a chord symbol, and add the information into "result". 
  // "symbol" is the original chord symbol with the front and middle stuff removed.
  function parseBackStuff(result, symbol) {
      var regex = RegExp(backStuff(),'g');
      var tokens = regex.exec(symbol);
      if (tokens && tokens[1]) {
          result.bass = {letter: tokens[1], sharp: tokens[2] || undefined, flat: tokens[3] || undefined};
      }
      return {result: result, symbol: regex.lastIndex ? symbol.substring(regex.lastIndex) : ""};
  }

  // Given a string representing a chord (e.g. "C#ma7b9"), returns a chord specification.
  function parseChordSymbol(symbol) {
      var fullSymbol = symbol;

      // If the entire symbol is enclosed in parentheses, we simply remove the parentheses.
      tokens = symbol.match(/^\((.*)\)$/);
      if (tokens) {
          symbol = tokens[1];
      }

      // If the chord symbol contains 1 or more substrings enclosed in parentheses,
      // remove those strings from the symbol and save them in the "extras" array.
      // We will deal with these in a moment.
      var extras = [];
      while (true) {
          tokens = symbol.match(/\((.*?)\)/);
          if (!tokens) break;
          extras.push(tokens[1]);
          symbol = symbol.replace(/\((.*?)\)/, "");
      }

      var result = {sus: [], add: [], drop: [], alter: []};

      // Do the front stuff
      var regex = RegExp(frontStuff(), "g");
      var tokens = regex.exec(symbol);
      if (tokens) {
          if (tokens[1]) result.letter = tokens[1];
          if (tokens[2]) result.sharp = true;
          if (tokens[3]) result.flat = true;
          if (tokens[4]) result.major = true;
          if (tokens[5]) result.minor = true;
          if (tokens[6]) result.diminished = true;
          if (tokens[7]) result.halfdim = true;
          if (tokens[8]) result.augmented = true;
          if (tokens[9]) result.triangle = true;
          if (tokens[10]) doNum(result, tokens[10], false);

          // trim the front stuff off the symbol
          symbol = symbol.substring(regex.lastIndex);
      }

      // do the middle stuff
      var temp = parseMiddleStuff(result, symbol);
      result = temp.result;
      symbol = temp.symbol;

      // do the back stuff
      if (symbol.length > 0) {
          temp = parseBackStuff(result, symbol);
          result = temp.result;
          symbol = temp.symbol;
          if (symbol.length > 0) {
              console.log("warning: we didn't fully parse the chord symbol", fullSymbol);
          }
      }

      // do the extras. Each extra contains "middle stuff".
      for (var i = 0; i < extras.length; i += 1) {
          temp = parseMiddleStuff(result, extras[i]);
          result = temp.result;
          if (temp.symbol.length !== 0) {
              console.log("warning: we didn't fully parse the chord symbol", fullSymbol);
          }
      }        

      // Uncomment the following line to see what the chord specification looks like.
      // console.log(symbol, JSON.stringify(result));

      return result;
  }

  // Given a chord specification, return an chordMap object containing all the notes of the chord.
  function expandChordSpec(chordSpec) {
      var result;
      var seventh;
      if (chordSpec.minor) {
          result = {1: 0, 3: -1, 5: 0};
          seventh = -1;

      } else if (chordSpec.diminished) {
          result = {1: 0, 3: -1, 5: -1};
          seventh = -2;
          result[7] = seventh;

      } else if (chordSpec.halfdim) {
          result = {1: 0, 3: -1, 5: -1};
          seventh = -1;
          result[7] = seventh;

      } else if (chordSpec.augmented) {
          result = {1: 0, 3: 0, 5: 1};
          seventh = -1;

      } else { // default is major
          result = {1: 0, 3: 0, 5: 0};
          seventh = chordSpec.major ? 0 : -1;
      }
      
      if (chordSpec.triangle) {
          seventh = 0;
          result[7] = seventh;
      }

      if (chordSpec.sixnine) {
          result[6] = 0;
          result[9] = 0;
      }

      var i;
      for (i = 0; i < chordSpec.drop.length; i += 1) {
          delete result[parseInt(chordSpec.drop[i])];
      }

      if (chordSpec.majoralt) {
          seventh = 0;
          chordSpec.number = chordSpec.majoralt;
      }

      if (chordSpec.number) {
          switch(chordSpec.number) {
              case "13": result[13] = 0;
              case "11": result[11] = 0;
              case  "9": result[9] = 0;
              case  "7": result[7] = seventh;
                         break;
              case "6":  result[6] = 0;
                         break;
              case "5":  delete result[3];
                         break;
          }
      }

      for (i = 0; i < chordSpec.sus.length; i += 1) {
          result[parseInt(chordSpec.sus[i])] = 0;
          delete result[3];
      }

      for (i = 0; i < chordSpec.add.length; i += 1) {
          result[parseInt(chordSpec.add[i])] = 0;
      }

      if (chordSpec.alt) {
          // "alt" is intended to be interpreted by the performer. Here we use 7#5#9 because that is a common choice.
          result[7] = -1;            
          result[5] = +1;
          result[9] = +1;
      }

      for(var i = 0; i < chordSpec.alter.length; i += 1) {
          var a = chordSpec.alter[i];
          result[parseInt(a.number)] = a.sharp ? 1 : -1;
      }

      return result;
  }

  // Deletes items from a chordMap object, to make it contain no more than maxLength notes.
  // Eliminates notes that are less important harmonically, in this order: tonic, perfect 5,
  // then secondary color notes (e.g. the 11th and 9th in a 13th chord).
  function prune(maxLength, chordMap) {
      var initialLength = countProperties(chordMap);

      // Keep deleting items until we have the right number of items.
      loop: while (countProperties(chordMap) > maxLength) {
          // delete the tonic note if it exists
          if (chordMap[1] != undefined) {
              delete chordMap[1];
              continue loop;
          }

          // delete the perfect 5 if it exists
          if (chordMap[5] == 0) {
              delete chordMap[5];
              continue loop;
          }

          // delete 2nd highest note, so e.g. we'll keep 13 but delete 11 and 9
          var indices = [];
          for (var i in chordMap) indices.push(i);
          delete chordMap[indices[indices.length - 2]];
      }
  }

  // Given a chordMap object, returns the corresponding array of midi notes,
  // based on the root note from "chordSpec".
  function render(chordMap, chordSpec) {
      var midiNotes = [];

      // If this chord has no letter (e.g. "/B"), use the previous chord
      if (!chordSpec.letter) {
          var prevSpec = render.lastChordSpec || {letter: "C"};
          chordSpec.letter = prevSpec.letter;
          chordSpec.sharp = prevSpec.sharp;
          chordSpec.flat = prevSpec.flat;
      }

      // Save the current chord in a static variable, in case we need it next time.
      render.lastChordSpec = chordSpec;

      // Find the tonic midi for this chord.
      var tonicMidiNote = letterToMidiNote(chordSpec);

      // For each note in chordMap, determine the corresponding midi note.
      var notesemitones = {1: 0, 2: 2, 3: 4, 4: 5, 5: 7, 6: 9, 7: 11, 8: 12, 9: 14, 10: 16, 11: 17, 12: 19, 13: 21};
      for (var i in chordMap) {
          // the midi note = tonic midi note + the number of semitones above the tonic + adjustment (-1,0,+1).
          midiNotes.push(tonicMidiNote + notesemitones[i] + chordMap[i]);
      }

      return midiNotes;
  }

  // Given an array of midiNotes, adjust the notes up or down by octaves, in order to find
  // the inversion in which exactly one note is higher than middle C.
  function findOptimumInversion(midiNotes, chordSpec) {
      function compare(a, b) {return a - b;}

      // To get the best sounding inversion, we need to use a slightly different threshold,
      // depending on the chord root. The threshold varies from middle C to Eb above middle C.
      var root = letterToInterval(chordSpec);
      var threshold = {0:60, 1:61, 2:62, 3:62, 4:63, 5:60, 6:62, 7:61, 8:60, 9:62, 10:62, 11:62}[root];

      switch(numNotesAboveThreshold(midiNotes, threshold)) {
          case 0:
              while(numNotesAboveThreshold(midiNotes, threshold) < 1) {
                  midiNotes[0] += 12;
                  midiNotes.sort(compare);
              }
          case 1:
              return;
          default:
              while(numNotesAboveThreshold(midiNotes, threshold) > 1) {
                  midiNotes[midiNotes.length -1] -= 12;
                  midiNotes.sort(compare);
              }
              return;
      }
  }

  // To an array of midi notes, add the bass note defined by the chord specification.
  function addBass(chordSpec, midiNotes) {
      var bassNote = letterToMidiNote(chordSpec.bass || chordSpec);
      if (bassNote >= 52) bassNote -= 12; // Maybe adjust the octave. We can go as low as E below the bass clef.
      if (bassNote != midiNotes[0]) midiNotes.unshift(bassNote);
  }

  // Search the current score for all the chord symbols in all tracks.
  // Returns an array of score chord symbol objects like {tick: 1440, duration: 960, text: "Db7"}.
  function findAllChordSymbols() {
      var chords = {};
      var segment = curScore.firstSegment();
      while (segment) {
          var annotations = segment.annotations; 
          for (var a in annotations) {
              var annotation = annotations[a];
              if (annotation.name == "Harmony") {
                  // Save the chord for this tick. If multiple tracks have a chord at the same tick,
                  // we will only keep the last one we find.
                  chords[segment.tick] = {tick: segment.tick, text: annotation.text};
              }
          }
          segment = segment.next;
      }

      // Calculate the duration of each chord = start time of next chord - start time of this chord.
      // Also, copy all the chords to an Array, we no longer need them to be in an Object.
      var result = [];
      for (var i in chords) {
          var chord = chords[i];
          if (result.length > 0) {
              result[result.length - 1].duration = chord.tick - result[result.length - 1].tick;
          }
          result.push(chord);
      }
      if (result.length > 0) {
          var scoreDuration = curScore.lastSegment.tick + 1;
          var lastItem = result[result.length - 1];
          lastItem.duration = scoreDuration - lastItem.tick;
      }
      return result;
  }

  // Given a text chord symbol (e.g. "C7b9"), return an array of midi notes (e.g. [60, 64, 67, 70, 73]).
  // If raw = true, you get all the notes of the chord, which can be 9 or 10 notes for complex chords.
  // If raw = false, you get at most 5 notes, in an inversion that puts most of the notes near or below middle C.
  function chordTextToMidiNotes(text, raw) {
      var chordSpec = parseChordSymbol(text);
      var chordMap = expandChordSpec(chordSpec);
      if (!raw) prune(4, chordMap);
      var midiNotes = render(chordMap, chordSpec);
      if (!raw) findOptimumInversion(midiNotes, chordSpec);
      addBass(chordSpec, midiNotes);
      return midiNotes;
  }

  // Given a rhythm pattern that is being repeated over the entire score,
  // create a slice of the pattern that corresponds to a given start time and duration.
  // The slice may span multiple iterations of the pattern.
  function sliceOfRhythmPattern(pattern, startTick, duration) {
      var needToFill = duration; // how much time we need to fill
      var prevItem = null;
      var ticksFilled = 0; // how much time we've filled so far

      // The duration of the pattern, in ticks. We use "reduce" to add together the duration
      // of all the items in the pattern.
      var patternDuration = pattern.reduce(function(acc,item) {return acc + item.duration;}, 0);
      var posInPattern = startTick % patternDuration; // where in the pattern we start at
      var item;
      var result = [];

      function addOneResultItem(item, overshoot) {
          var useDuration = Math.min(needToFill, item.duration - overshoot);
          result.push({tick: ticksFilled + startTick, voicing: item.voicing, duration: useDuration});
          posInPattern += useDuration;
          ticksFilled += useDuration;
          needToFill -= useDuration;
          prevItem = item;
      }

      var i = 0;
      while (needToFill > 0) {
          item = pattern[i % pattern.length];
          var itemTick = item.tick + (Math.floor(i / pattern.length) * patternDuration);

          if (itemTick < posInPattern) {
              prevItem = item;
              i += 1;

          } else if (itemTick === posInPattern) {
              addOneResultItem(item, 0);
              i += 1;

          } else if (itemTick > posInPattern) {
              addOneResultItem(prevItem, posInPattern - prevItem.tick);
          }
      }
      return result;
  }

  // Write a bunch of chords to a given track of the current score, starting at time 0.
  // "chords" is an array of score chord symbol objects like {tick: 1234, text: "Db7", duration: 234}.
  // "theRhythm" is an optional array which describes the rhythm to use for each chord. 
  // If theRhythm is not given, we generate a single chord for each chord.
  function writeChords(chords, track, theRhythm, options) {
      if (chords.length == 0) return;

      clearStaff(track);

      var cursor = curScore.newCursor();
      cursor.track = track;
      cursor.rewind(Cursor.SCORE_START);

      // Move the cursor to the first chord. NOTE: this code doesn't always work perfectly, it fails to
      // position the cursor exactly at the time of the first chord. I don't know why.
      while (cursor.tick <= chords[0].tick) {
          cursor.next();
      }
      cursor.prev();

      // Process each chord
      for (var i in chords) {
          var theChord = chords[i]; // the chord we're working on

          // find the notes we need to write for this chord
          var midiNotes = chordTextToMidiNotes(theChord.text, options.raw);

          // find the rhythm
          var rhythm;
          if (theRhythm && theRhythm.length > 0) {
              // the rhythm was given to us. 

              if (options.useEntirePattern) {
                  // In this mode, we keep repeating the rhythm pattern over and over again
                  // for the entire score. Right now, find the slice of the pattern
                  // that corresponds to the current chord's start time and duration.
                  rhythm = sliceOfRhythmPattern(theRhythm, theChord.tick, theChord.duration);

              } else {
                  // Make a copy of the rhythm and make the ticks
                  // be relative to the beginning of the score.
                  rhythm = [];
                  var rhythmDuration = 0;
                  for (var n = 0; n < theRhythm.length; n += 1) {
                      var item = shallowCopy(theRhythm[n]);
                      item.tick += theChord.tick;
                      rhythmDuration += item.duration;
                      rhythm.push(item);
                      if (rhythmDuration >= theChord.duration) {
                          rhythm[rhythm.length - 1].duration -= rhythmDuration - theChord.duration;
                          break;
                      }
                  }
                  if (rhythmDuration < theChord.duration) {
                      rhythm[rhythm.length - 1].duration += theChord.duration - rhythmDuration;
                  }
              }
          } else {
              // the rhythm was not provided. We'll just do one chord for the whole duration.
              rhythm = [theChord];
          }

          // Loop over the notes in the rhythm pattern. Generate the midiNotes for each note in the pattern.
          for (var k = 0; k < rhythm.length; k += 1) {
              var chord = rhythm[k];

              // find the chord's duration. Adjust the duration if the cursor is not exactly at the desired time.
              var duration = chord.duration + (chord.tick - cursor.tick); // in midi ticks
              var bumpCount = 0; // see below for explanation

              // If the chord's duration is long, we will write several shorter notes, rather than
              // one long note. We will loop until the entire duration has been used up.
              while (duration > 0) {
                  // Find out when the current measure ends. This is awkward code, do you know a better way?
                  var endOfThisMeasure = cursor.measure.nextMeasure
                      ? cursor.measure.nextMeasure.firstSegment.tick
                      : cursor.tick + duration;

                  // Set the note's duration so that the note ends no later than the end of the measure.
                  var thisDuration = Math.min(duration, endOfThisMeasure - cursor.tick);
                  duration -= thisDuration;
                  cursor.setDuration(thisDuration / 60, 32);

                  // Add all the midi notes to the score.
                  var beforeTick = cursor.tick; 
                  if (chord.voicing === "rest") {
                      addRest(cursor);
                  } else {
                      var gotNote = false;
                      if (chord.voicing !== "nonbass") {
                          cursor.addNote(midiNotes[0], gotNote);
                          gotNote = true;
                      }
                      if (chord.voicing !== "bass") {
                          for (var j = 1; j < midiNotes.length; j += 1) {
                              cursor.addNote(midiNotes[j], gotNote);
                              gotNote = true;
                          }
                      }
                  }

                  // NOTE: there is a limitation with cursor.addNote(), it sometimes adds a note shorter
                  // than the requested duration. The following workaround adds the missing time 
                  // back into "duration", and then continues looping.
                  // TODO: find a way to tie all these notes together.
                  var actualDuration = cursor.tick - beforeTick;
                  if ((actualDuration != thisDuration) && cursor.measure.nextMeasure) {
                      duration += thisDuration - actualDuration;
                      console.log("** at time", cursor.tick, "didn't get requested duration. Wanted", thisDuration,
                          "got", actualDuration, "Bumping duration", thisDuration - actualDuration, "to", duration);
                      if (++bumpCount > 3) {
                          console.log("bailout!!"); // avoid an infinite loop in case this workaround goes awry
                          duration = 0;
                      }
                  }
              }
          }
      }
  }

  // Here is where do all the work. It's easy - we find all the chords, then write them to the score.
  function expandChordSymbols(pattern, options) {
      curScore.startCmd();
      if (pattern) savePattern(pattern);
      writeChords(findAllChordSymbols(), curScore.ntracks - 4, pattern, options);
      curScore.endCmd();
  }
/**/
  property bool working:false
  Component { id: msg ; MessageDialog { title: "Confirm"; standardButtons: StandardButton.Ok; onAccepted: console.log("ok") } }
  function runapp(){
    if(!curScore) msg.createObject(pluginscope,{ 
      text:"The action cannot be completed because no score file is opened", visible: true})
    if(!working && curScore){
      working=true
      curScore.startCmd();
      runsheet();
    }
  }
  function finapp(exit){
    if(working && curScore) curScore.endCmd();
    working=false
    if(exit) plugindialog.visible=false //DOCKED        //REPEATDELTILEOF
  }
/**/
  // option array ordering = visual, is coupled with action for sake of simplicity, settings.[option][n]
  property var halfmodel:[
       [keySignEdit      ,
        '<b>Use key (for I-VII only) :&nbsp;&nbsp;&nbsp;</b>'
        +getNoteName(settings.keysig+7+(settings.keymM===0?3:0))
        +(settings.keymM===0?' minor ':' Major ')                         ,'keyedit'              ,["-", "Edit"] ]
      ,[existingsym      ,'<b>Existing symbols:</b>'                      ,'existingsym'          ,["Ignore","Ignore, add brackets","Remove","Use (Chord symbols only)"] ]
      ,[durationMode     ,'<b>Pedal tones are chord tones :</b>'          ,'entire_note_duration' ,["No","Yes"] ]
      ,[symbolMode       ,'<b>Identify as :</b>'                          ,'symformat'            ,["None","I-VII <sup>6</sup>,<sup>6</sup><sub>4</sub> ...", "I-VII a,b,c,d", "I-VII","Chord <sup>1,2,3 ...</sup>","Chord / Bass","Chord"] ]
      ,[symPos           ,'<b>Symbol placement :</b>'                     ,'sympos'               ,["Bottom","Below selected","Above selected","Top"] ]
      ,[keyTonality      ,'<b>Use key (for I-VII only) :&nbsp;&nbsp;</b>' ,'keymM'                ,["minor","Major"] ]
      ,[keySigns         ,undefined                                       ,'keysig'               ,["7b","6b","5b","4b","3b","2b","1b","None","1#","2#","3#","4#","5#","6#","7#"] ]
      ,[rnainvobject     ,'<b>Object for inversion (for I-VII only) :</b>','rnainvobject'         ,["Figured bass object","Staff Text object","Default"]]
      ,[engineEdit       ,'<b>Advanced (for Chord only) :</b>'            ,'engedit'              ,["Hide","Show"]]
      // ,[romOpHead        ,'<b>For I-VII :</b>'                         ,undefined              ,[] ]
      // ,[rom24            ,'<b>54 or 52 :</b>'                          ,'rom24'                ,['Disabled','52','54'] ]
      // ,[rom74            ,'<b>754 :</b>'                               ,'rom74'                ,['Disabled','Enabled'] ]
      // ,[jazzOpHead       ,'<b>For A-G :</b>'                           ,undefined              ,[] ]
      ,[jazz6            ,'<b>6 or m7 :</b>'                              ,'jazz6'                ,['m7','6'] ]
      ,[jazzm6           ,'<b>m6 or m7b5 :</b>'                           ,'jazzm6'               ,['crossed circle 7', 'm7b5','m6'] ]
      ,[jazz69           ,'<b>69 or m7(11) :</b>'                         ,'jazz69'               ,['m7(11)','69'] ]
      ,[jazz24           ,'<b>sus2, sus4 :</b>'                           ,'jazz24'               ,['Disabled','Sus4 only','Both, prefer sus2'] ]
      ,[showrepeat       ,'<b>Show unchanged harmony :</b>'               ,'showrepeat'           ,["None","Mark with -","Repeat the symbol"] ]
      ,[partialChordMode ,'<b>Show incomplete chords :</b>'               ,'partialChords2'       ,["None","Mark with X","Closest match"] ]
      ,[listnotesMode    ,'<b>List tones :</b>'                           ,'listnotes'            ,["None","Incomplete chords","All"] ]
      ,[chordColorMode   ,'<b>Color tones by interval :</b>'              ,'colortones'           ,["No","Yes"] ]
      ,[runandexitMode   ,'<b>Close window after completion :</b>'        ,'runandexit'           ,["No","Yes"] ]//DOCKED
    ]
  function unidirdflow(mi,ni){ settings[halfmodel[Number(mi)][2]]=Number(ni) }
  function hexToHSL(hex) {
    var result = /^#?([a-f\d]{2})([a-f\d]{2})([a-f\d]{2})$/i.exec(hex),
      r = parseInt(result[1], 16),
      g = parseInt(result[2], 16),
      b = parseInt(result[3], 16);
      r /= 255, g /= 255, b /= 255;
      var max = Math.max(r, g, b), min = Math.min(r, g, b);
      var h, s, l = (max + min) / 2;
      if(max == min){
        h = s = 0; // achromatic
      }else{
        var d = max - min;
        s = l > 0.5 ? d / (2 - max - min) : d / (max + min);
        switch(max){
          case r: h = (g - b) / d + (g < b ? 6 : 0); break;
          case g: h = (b - r) / d + 2; break;
          case b: h = (r - g) / d + 4; break;
        }
        h /= 6;
      }
    return [h,s,l].map(function(x){return x.toFixed(2)}).join(', ')
  }
  function shallowcopy(xs){var ys=[];for(var i in xs) if(typeof xs[i]!='function') ys[i]=xs[i];return ys}
  function reverseNonmutate(xs){var ys=shallowcopy(xs);ys=ys.reverse();return ys}
  function ksstr(ni,minor){return getNoteName(ni+7+(minor?3:0))+(minor?'m':'')}
  //qmllearning dialog boilerplate 20230227A  https://musescore.org/en/node/345720
  property real wrapperwidth: 600
  property bool fixsize:false
  property int  pagemargintb:10; property int pagemarginlr:20
  property real customwidth: Screen.desktopAvailableWidth*.5
  property real customheight: Screen.desktopAvailableHeight*.8
  onRun:{
    if(fixsize){
      plugindialog.width =Math.min(customwidth  ,Screen.desktopAvailableWidth)
      plugindialog.height=Math.min(customheight ,Screen.desktopAvailableWidth)
    }else{
      // plugindialog doesn't auto expand by contentItem : qml bug ?
      plugindialog.width =Math.min( Math.max(wrapperwidth, plugincontent.width) ,Screen.desktopAvailableWidth)
      // height is tricky
      setTimeout.createObject(pluginscope,{ interval:1 ,running:true  }).triggered.connect(function(){ 
        plugindialog.height=
          Math.min( 
            grid.height+plugincontent.header.height
            +allcontent.Layout.topMargin+allcontent.Layout.bottomMargin 
          , Screen.desktopAvailableHeight)
      })
    }
  }//DOCKED
  property bool ctrl:false
  function handlepress(event){
    if (event.key === Qt.Key_Return || event.key == Qt.Key_Enter) runapp()
    if (event.key === Qt.Key_Escape) finapp(true)//DOCKED
    if (event.key === Qt.Key_Control ) ctrl=true
    if (event.key === Qt.Key_Z && ctrl) cmd('undo')
  }
  function handlerelease(event){
    if (event.key === Qt.Key_Control ) ctrl=false
  }
  Dialog { id:plugindialog//DOCKED
    title: pname//DOCKED
    visible:true//DOCKED
    //
      property var itemheight:buttonCancel.height
      property var itemwidth:buttonCancel.width
      //label and radio
        Component { id: headSpacingTmp; Rectangle{} }
        Component { id: headTmp; Text {textFormat: Text.RichText; height: plugindialog.itemheight; verticalAlignment: Text.AlignVCenter} }//DOCKED
        Component { id: rbTmp; RadioButton { height: plugindialog.itemheight;//DOCKED
          indicator: null
          background:Rectangle{
            color: parent.activeFocus ? "#f0f6ff" : "transparent"
            anchors.fill:parent
          }
          contentItem: Text {
            font.bold: checked 
            textFormat: Text.RichText
            text: parent.text
            verticalAlignment: Text.AlignVCenter
          }
          property int mi; property int v; onClicked:{unidirdflow(mi,v)}
          property string tips:''; ToolTip.visible: tips!=''&&hovered
          ToolTip.text: tips
          // onPressed: stealfocus.forceActiveFocus() //don't show background if use mouse
          Keys.onPressed:handlepress(event)
          Keys.onReleased:handlerelease(event)
        }}
      //color
        ColorDialog {  id: colorDialog
          property string settingname
          onAccepted: { settings[settingname]=colorDialog.color }
        }
        Component { id: cpTmp; Button {
          id:btn
          height:40
          property var widthx:2
          width: this.height * widthx
          property string btnstr
          property string tipstr:''
          property string curcolor
          ToolTip.visible: hovered
          ToolTip.text:(tipstr!=''?tipstr+'\n':'')+'HEX: '+curcolor+'\nHSL: '+hexToHSL(curcolor)
          contentItem:Item{ anchors.fill: parent
            Row{
              Text{
                text: btnstr
                width: btn.contentItem.width - rect.width
                height: btn.height
                horizontalAlignment: Text.AlignHCenter; verticalAlignment: Text.AlignVCenter }
              Rectangle {
                id:rect
                color: curcolor
                width: btn.height
                height: btn.height}
              }
            }
          Keys.onPressed:handlepress(event)
          Keys.onReleased:handlerelease(event)
        }}
    //init and style
    Component.onCompleted:{//function initandstyle(){//DOCKED
      var radios=[],labels=[]
      var findlinked=function(obj,id){return obj[halfmodel.map(function(m,mi){return m[0]===id?mi:false}).filter(function(x){return x!==false})[0]]}
      //
      settings.keyedit=0
      settings.engedit=0
      halfmodel.map(function(m,mi){radios[mi]=[];
        var c=m[0],t=m[1]
        // if(c===keyTonality)
        //   headTmp.createObject(c,{width:findlinked(labels,keySigns).width+10})
        // if(~[/*romOpHead,*/jazzOpHead].indexOf(c))
        //   headTmp.createObject(c,{width:Qt.binding(function(){return Math.max(5,innercontent.width*.02)}) })
        if(~[/*rom24,rom74,*/jazz6,jazzm6,jazz69,jazz24].indexOf(c))
          headTmp.createObject(c,{width:Qt.binding(function(){return Math.max(10,innercontent.width*.05)}) })
        var label=headTmp.createObject(c,{
          text:Qt.binding(function(){return(
             (t?t:'')
            +(c===keyTonality?' '+getNoteName(settings.keysig+7+(settings.keymM===0?3:0))+' ':'')   )})   })
        labels[mi]=label
        reverseNonmutate(m[3]) // align 0/ option "No" to the right, see halfmodel's array
          .map(function(t,ni){
            var v=m[3].length-1-ni; // reverse back
            radios[mi][v]=rbTmp.createObject(c,{
               mi:mi, v:v, text:t, checked:settings[m[2]]===v
              ,tips: c===keySigns?ksstr(v)+'\n'+ksstr(v,true):''
              ,leftPadding:c===keySigns?5:10
              ,rightPadding:c===keySigns?5:10       })   })
        //reflow radio
        c.width=Qt.binding(function(){return c.parent.width})
      })
      findlinked(radios,symbolMode)[4].visible=false //disable superscript
      //advanced
      // romOp  .width=Qt.binding(function(){return  romOp  .parent.width  })
      jazzOp .width=Qt.binding(function(){return  jazzOp .parent.width  })
      //redundant
      findlinked(radios,keySigns)[0].visible=false
      findlinked(radios,keySigns)[14].visible=false
      //key edit
      keySignEdit.visible=Qt.binding(function(){return settings.keyedit==0})
      keyTonality.visible=Qt.binding(function(){return settings.keyedit==1})
      keySigns.visible=Qt.binding(function(){return settings.keyedit==1})
      findlinked(radios,keySignEdit)[0].visible=false
      //engine edit
      // romOp.visible=Qt.binding(function(){return settings.engedit==1})
      jazzOp.visible=Qt.binding(function(){return settings.engedit==1})
      findlinked(radios,engineEdit)[0].visible=Qt.binding(function(){return settings.engedit==1})
      findlinked(radios,engineEdit)[1].visible=Qt.binding(function(){return settings.engedit==0})
      //reflow colorpickers
      ;[symColorEdit,partialChordColorEdit,noteColorEdit].map(function(c){
        c.width=Qt.binding(function(){return c.parent.width}) 
        c.topPadding=10;c.bottomPadding=10
        if(c===noteColorEdit) c.spacing=10      })
      //
      ;[ [symColorEdit,'symColor']
        ,[partialChordColorEdit,'partialChordsColor']
      ].map(function(n){
        cpTmp.createObject(n[0],{
           curcolor: Qt.binding(function(){return settings[n[1]] })
          ,btnstr: "Color"
          ,widthx:2.5
        }).clicked.connect(function(){
            colorDialog.settingname=n[1]
            // colorDialog.currentColor=settings[n[0]]
            colorDialog.title="Color of the closest match chord symbol"
            colorDialog.visible=true 
        })
      })
      //
      ;[ ['colorR','Root']
        ,['color3','3rd\nsus4']
        ,['color5','5th']
        ,['color7','7th\nbb7th']
        ,['color9','9th\n2nd']
        ,['color11','11th\n#4th']
        ,['color13','13th\n6th']
        ,['colorNCT','Non Chord Tone','NCT']
      ].map(function(n){
        cpTmp.createObject(noteColorEdit,{
           curcolor: Qt.binding(function(){return settings[n[0]]})
          ,btnstr: n[2]||n[1]
          ,tipstr: n[1]
        }).clicked.connect(function(){
            colorDialog.settingname=n[0]
            // colorDialog.currentColor=settings[n[0]]
            colorDialog.title="Color of "+n[1]+"s"
            ;colorDialog.visible=true
        })
      })
    }
    //
    contentItem:Page{ id:plugincontent//DOCKED
    // Page{ id:plugincontent //DOCKED
    //   anchors.fill: parent //DOCKED
      header: ToolBar { //  ToolBar, TabBar, or DialogButtonBox 
        background: Rectangle { color: "#e3e3e3"}
        RowLayout{
          anchors.fill: parent
          ProgressBar { value: 0; id:pbar
            Layout.fillWidth: true
            Layout.leftMargin: plugindialog.itemwidth*.2 ; Layout.rightMargin: plugindialog.itemwidth*.2//DOCKED
            height: plugindialog.itemheight//DOCKED
          }
          ToolButton { id: buttonOK
            text: working?"    Running    ":"    Analyse    "
            onClicked: runapp()
            Keys.onPressed:handlepress(event)
            Keys.onReleased:handlerelease(event)
          }
          ToolButton { id: buttonCancel
            text: "    Cancel    "
            onClicked: finapp(true)
            Keys.onPressed:handlepress(event)
            Keys.onReleased:handlerelease(event)
          }
        }
      }
      contentItem:Rectangle{ id:pagecontent
        color: "#e3e3e3"
        Flickable{
          anchors.fill:parent;
          contentWidth: grid.width; contentHeight: grid.height
          GridLayout{ id:grid
            columns: 1
            Column{ id:allcontent
              Layout.preferredWidth: pagecontent.width-Layout.leftMargin-Layout.rightMargin 
              Layout.leftMargin : pagemarginlr; Layout.rightMargin  :pagemarginlr
              Layout.topMargin  : pagemargintb; Layout.bottomMargin :pagemargintb
              Column{ id:innercontent
                width: parent.width
                //
                Flow{ id: existingsym }
                Flow{ id: durationMode }
                Flow{ id: symbolMode }
                Flow{ id: engineEdit }
                  // Column{id: romOp
                  //   Flow{ id:romOpHead }
                  //   Flow{ id: rom24  }
                  //   Flow{ id: rom74  }
                  // }
                  Column{id: jazzOp
                    // Flow{ id:jazzOpHead }
                    Flow{ id: jazz6 }
                    Flow{ id: jazzm6 }
                    Flow{ id: jazz69 }
                    Flow{ id: jazz24 }
                  }
                Flow{ id: keySignEdit }
                  Flow{ id: keyTonality }
                  Flow{ id: keySigns }
                Flow{ id: rnainvobject }
                Flow{ id: symPos }
                Flow{ id: symColorEdit }
                Flow{ id: partialChordMode }
                Flow{ id: partialChordColorEdit }
                Flow{ id: showrepeat }
                Flow{ id: listnotesMode }
                Flow{ id: chordColorMode }
                Flow{ id: noteColorEdit }
                Flow{ id: runandexitMode } //DOCKED
              }
            }
          }
        }
        Item { id:stealfocus
          focus:true
          Keys.onPressed:handlepress(event)
          Keys.onReleased:handlerelease(event)
        }
      }
    }
  }//DOCKED
}

//REPEAT
/*
    plugindialog.visible=false
  }
  onRun:{
    plugindialog.width =Math.min( Screen.desktopAvailableWidth*.3 ,Screen.desktopAvailableWidth)
    plugindialog.height=plugindialog.itemheight
    setTimeout.createObject(pluginscope,{ interval:1 ,running:true  }).triggered.connect(function(){ 
      runapp() })
  }
  function handlepress(event){
    if (event.key === Qt.Key_Escape) finapp(true)
  }
  Dialog { id:plugindialog
    title: pname
    visible:true
    property var itemheight:buttonCancel.height
    property var itemwidth:buttonCancel.width
    //
    contentItem:ToolBar {
      Item { id:stealfocus
        focus:true
        Keys.onPressed:handlepress(event)
      }
      anchors.fill: parent
      background: Rectangle { color: "#e3e3e3"}
      RowLayout{
        width: parent.width
        ProgressBar { value: 0; id:pbar
          Layout.fillWidth: true
          Layout.leftMargin: plugindialog.itemwidth*.2 ; Layout.rightMargin: plugindialog.itemwidth*.2
          height: plugindialog.itemheight
        }
        ToolButton { id: buttonCancel
          text: "    Cancel    "
          onClicked: finapp(true)
          Keys.onPressed:handlepress(event)
        }
      }
    }
  }
}
*/